import pandas as pd

def create_quantile_alphas(instances: pd.DataFrame, prediction_result: pd.DataFrame, quantile_threshold) -> pd.DataFrame:
    """
    - based on prediction result and quantile threshold (symeteric), creates +1, 0, -1 signals
    - returns columns with Prediction (the same as input) + Signal (+1, 0, -1)
    """

    assert len(prediction_result.columns) == 1, "prediction has more than one column"
    assert 0 < quantile_threshold < 1, f"quantiles need to be between 0 and 1, provided: {quantile_threshold}"

    alphas = prediction_result.copy()
    alphas.columns = ["Prediction"]

    alphas = alphas[alphas.index.isin(prediction_result.index)]
    alphas = pd.concat([alphas, instances], axis=1)

    # todo: just dropped na
    alphas = alphas.dropna()

    alphas["Signal"] = 0
    quant_low = quantile_threshold
    quant_high = 1 - quantile_threshold

    low = alphas["Prediction"].quantile(quant_low)
    high = alphas["Prediction"].quantile(quant_high)

    alphas.loc[alphas["Prediction"] < low, "Signal"] = -1
    alphas.loc[alphas["Prediction"] > high, "Signal"] = 1

    return alphas

def create_absolute_prediction_alphas(instances: pd.DataFrame, prediction_result: pd.DataFrame) -> pd.DataFrame:
    """
    - use the prediction as alpha signal
    """
    assert len(prediction_result.columns) == 1, "prediction has more than one column"

    alphas = prediction_result.copy()
    alphas.columns = ["Prediction"]

    alphas = alphas[alphas.index.isin(prediction_result.index)]
    alphas = pd.concat([alphas, instances], axis=1)

    # todo: just dropped na
    alphas = alphas.dropna()

    alphas["Signal"] = alphas["Prediction"]
    return alphas


import pytest
import numpy as np
import pandas as pd

from moosir_feature.transformers.basic_features.feature_calculator import *


# @pytest.fixture
# def ohlc_prices():
#     return pd.DataFrame(data=np.random.rand(10, 4), columns=["Open", "High", "Low", "Close"])


def test_create_time_to_markets_feature():
    sample_n = 3
    ohlc_prices = pd.DataFrame(data=np.random.rand(sample_n, 4),
                               columns=["Open", "High", "Low", "Close"],
                               index=pd.date_range("2010-10-10", periods=sample_n, freq="H"))

    # act
    result = calculate_features(ohlc=ohlc_prices, feature_names=["Market-Open-Close-Dist"])

    # assert
    # todo: assert
    print(result)


def test_create_day_of_week():
    sample_n = 3
    ohlc_prices = pd.DataFrame(data=np.random.rand(sample_n, 4),
                               columns=["Open", "High", "Low", "Close"],
                               index=pd.date_range("2022-03-02", periods=sample_n, freq="D"))

    # act
    result = calculate_features(ohlc=ohlc_prices, feature_names=["Dow"])

    # assert
    # todo: assert
    assert result["Dow"].values[0] == 2 # Wed
    assert result["Dow"].values[1] == 3 # Thur
    assert result["Dow"].values[2] == 4 # Fri



import logging
import numpy as np
import pandas as pd
from sklearn.utils import shuffle
from sklearn.model_selection import TimeSeriesSplit
from pandas.testing import assert_index_equal

TRAIN_FEATURE_KEY = "train_feature"
TRAIN_TARGET_KEY = "train_target"
TEST_FEATURE_KEY = "test_feature"
TEST_TARGET_KEY = "test_target"

logger = logging.getLogger(__file__)

def shuffle_df(data: pd.DataFrame, target_col: str, rand_state: int = None):
    targets = data[[target_col]]
    features = data[data.columns[~data.columns.isin([target_col])]]

    assert len(features.columns) > 1, "no column left for feature"

    if rand_state is None:
        rand_state = np.random.randint(1000)

    features_shuf_ind, targets_shuf_ind = shuffle(features.index, targets.index, random_state=rand_state)

    features_shuf = features.loc[features_shuf_ind]
    targets_shuf = targets.loc[targets_shuf_ind]

    all_shuf = data.loc[features_shuf_ind]

    return features_shuf, targets_shuf, all_shuf, rand_state


class CustomTsCv:
    """
      in addition to sklearn time series can shuffle train data in blocks
    """
    def __init__(self, train_n: int, test_n: int, sample_n: int, train_shuffle_block_size: int = None):
        split_n = int((sample_n + 1) / test_n) - 1
        print(split_n)
        assert split_n > 1, "split is 1"
        self.split_n = split_n
        self.train_n = train_n
        self.test_n = test_n
        self.sample_n = sample_n
        self.train_shuffle_block_size = train_shuffle_block_size


        if train_shuffle_block_size is not None:
            assert train_n % train_shuffle_block_size == 0, "train data is not devidable in shuffle block size"

        self.train_shuffle_block_size = train_shuffle_block_size
        self.tscv = TimeSeriesSplit(max_train_size=train_n, test_size=test_n, n_splits=split_n)

    def split(self, X, y, groups=None):
        counter = 0
        for tr, ts in self.tscv.split(X=X, y=y, groups=groups):
            if counter == 0:
                counter = counter + 1
                yield tr, tr # to be inline with cross_val_predict expected cv
                             # google: "cross_val_predict only works for partitions TimeSeriesSplit"

            final_tr = tr
            final_ts = ts

            if self.train_shuffle_block_size is not None:
                if len(final_tr) % self.train_shuffle_block_size == 0:
                    np.random.shuffle(final_tr.reshape((-1, self.train_shuffle_block_size)))
                else:
                    logger.warning(f"train shuffle cant be done in this itt: final_tr: {len(final_tr)}")



            yield final_tr, final_ts

    def get_n_splits(self, X, y, groups=None):
        return self.tscv.n_splits + 1 # it is for first split that train returns (check out counter in split function)


def calculate_confidence_interval(statistics: pd.DataFrame,
                                  alpha: float = 5.0,
                                  ignore_small_instances: bool = False) -> pd.DataFrame:
    """
    to calculate statistical intervals based on trials
    :param statistics: any estimations coming from running trials (e.g. accuracy, error, ...).
        - dataframe, columns: statistics, row: trials (samples)
    :param alpha: p-value
    :param ignore_small_instances: for too small values, this method cant work, if True, the answer might not be reliable
    :return: dataframe: columns: lower/higher ci, index: statistic name from input statistics
    """
    if not ignore_small_instances:
        assert len(statistics) < 100, "number of trials is too low to build any confidence level"

    # calculate lower percentile (e.g. 2.5)
    lower_p = alpha / 2.0
    # calculate upper percentile (e.g. 97.5)
    upper_p = (100 - alpha) + (alpha / 2.0)

    cis = statistics.quantile([lower_p / 100.0, upper_p / 100.0]).transpose()
    cis.columns = [f"Lower_{lower_p}", f"Upper_{upper_p}"]
    return cis


# class CvTsDfUtil:
#     @staticmethod
#     def get_split_len(fold_result: dict):
#         return len(fold_result[TEST_FEATURE_KEY])
#
#     @staticmethod
#     def get_train(fold_result: dict, ind: int):
#         return fold_result[TRAIN_FEATURE_KEY][ind], fold_result[TRAIN_TARGET_KEY][ind]
#
#     @staticmethod
#     def get_test(fold_result: dict, ind: int):
#         return fold_result[TEST_FEATURE_KEY][ind], fold_result[TEST_TARGET_KEY][ind]



import pandas as pd
import pytest
import numpy as np

from moosir_feature.transformers.managers.feature_cv_manager import FeatureConfidenceLevelManager
from moosir_feature.transformers.managers.feature_manager import *
from moosir_feature.transformers.features_cv.feature_cv_manager import *
from moosir_feature.transformers.tsfresh_features.feature_calculator import *
from moosir_feature.transformers.indicators.feature_calculator import *
from moosir_feature.transformers.basic_features.feature_calculator import *

# from src.research.preproc.pipeline_kedro.src.preproc.domain.transformers.tsfresh_features.contexts import *

SAMPLE_N = 200


@pytest.fixture
def ohlc():
    df = pd.DataFrame(data=np.random.rand(SAMPLE_N, 4),
                      columns=["Open", "High", "Low", "Close"],
                      index=pd.date_range(start="01/01/2000", periods=SAMPLE_N, freq="10T"),
                      )
    df.index.name = "Timestamp"
    return df


def _get_target_settings(setting_key: str):
    ts = TsfreshTargetSettings(**{
        'win_len': 80,
        'lookahead_len': 3,
        'feature_name': 'linear_trend_timewise',
        'target_col_feature_type': 'Trend',
        'target_col_feature_subtype': 'Slope',

    })

    ind = IndicatorTargetSettings(**{
        'win_len': 80,
        'lookahead_len': 3,
        'feature_name': RsiOperator.__name__,
        'target_col_feature_type': 'Rsi',
        'target_col_feature_subtype': RsiOperator.get_column_names()[0],
    })
    settings = {"ts": ts, "ind": ind}
    return settings[setting_key]


def _get_feature_settings(setting_key: str):
    ts = TsfreshFeatureSettings(**{
        'win_lens': [80],
        'feature_names': ["abs_energy"]})

    ind = IndicatorFeatureSettings(**{
        'win_lens': [80],
        'feature_names': [VilliamrOperator.__name__]
    })

    basic = BasicFeatureSettings(**{
        'feature_names': ["Hour"]
    })

    settings = {"ts": ts, "ind": ind, "basic": basic}
    return settings[setting_key]


def _get_lag_settings(setting_key: str):
    ts = TsfreshLagSettings(**{'feature_parameters': {
        'win_lens': [80],
        'feature_names': ["benford_correlation"]

    },
        'lag_lens': [3]
    })

    ind = IndicatorLagSettings(**{'feature_parameters':
        {
            'win_lens': [80],
            'feature_names': [AtrOperator.__name__]
        },
        'lag_lens': [3]
    })
    settings = {"ts": ts, "ind": ind}
    return settings[setting_key]


def _get_feature_selector_settings():
    return FeatureSelectorSettings(features_to_select_perc=0.3)


def _get_feature_cv_settings():
    return FeatureCVSettings(f_confidence_win_step_perc=0.2, f_confidence_win_len=100)


def _run_feature_cv(ohlc, target_settings, feature_settings_list, lag_settings_list):
    feature_selector_settings = _get_feature_selector_settings()
    cv_settings = _get_feature_cv_settings()

    fc_mgr = FeatureCreatorManager(target_settings=target_settings,
                                   feature_settings_list=feature_settings_list,
                                   lag_settings_list=lag_settings_list)

    fs_mgr = FeatureSelectorManager(feature_selector_settings=feature_selector_settings)

    fcv_mgr = FeatureConfidenceLevelManager(feature_cv_settings=cv_settings,
                                            feature_creator_mgr=fc_mgr,
                                            feature_selector_mgr=fs_mgr)
    # act
    features_cvs = fcv_mgr.calculate_cv(instances=ohlc)

    return features_cvs


def _get_indicator_operators(feature_names: list):
    return [OPERATORS_IND[f] for f in feature_names]


def _assert_tsfresh_features(features, feature_settings_list):
    features_df = _convert_features_csv_to_columns(features_cv=features)

    for f_settings in feature_settings_list:
        assert validate_columns(data=features_df,
                                feature_names=f_settings.feature_names,
                                win_lens=f_settings.win_lens)


def _assert_tsfresh_lags(features_cv, lag_settings_list):
    features_df = _convert_features_csv_to_columns(features_cv=features_cv)

    for l_settings in lag_settings_list:
        assert validate_columns(data=features_df,
                                feature_names=l_settings.feature_names,
                                win_lens=l_settings.win_lens)


def _assert_indicator_lags(features_cv, context, lag_settings_list):
    features_df = _convert_features_csv_to_columns(features_cv=features_cv)

    for l_settings in lag_settings_list:
        lag_ops = _get_indicator_operators(l_settings.feature_names)
        assert context.validate(features_df,
                                operators=lag_ops,
                                periods=l_settings.win_lens,
                                lag_periods=l_settings.lag_lens)


def _convert_features_csv_to_columns(features_cv):
    return pd.DataFrame(np.zeros([1, len(features_cv["Feature"].values)]), columns=features_cv["Feature"].values)


def _assert_indicator_features(context, features_cv, feature_settings_list):
    features_df = _convert_features_csv_to_columns(features_cv=features_cv)

    for f_settings in feature_settings_list:
        feature_ops = _get_indicator_operators(f_settings.feature_names)
        assert context.validate(features_df, operators=feature_ops, periods=f_settings.win_lens)


@pytest.mark.skip("debug takes too long and it fails")
def test_feature_manager_calculate_multiple_settings(ohlc: pd.DataFrame):
    # arrange
    target_settings = _get_target_settings("ts")
    feature_settings_list = [_get_feature_settings("ts"), _get_feature_settings("ind")]
    lag_settings_list = [_get_lag_settings("ts"), _get_lag_settings("ind")]

    # act
    features_cv = _run_feature_cv(ohlc=ohlc,
                                  target_settings=target_settings,
                                  feature_settings_list=feature_settings_list,
                                  lag_settings_list=lag_settings_list)

    # assert
    # todo: no target validation
    # assert features
    context = Context()
    _assert_tsfresh_features(features=features_cv, feature_settings_list=[feature_settings_list[0]])
    _assert_indicator_features(features_cv=features_cv, context=context,
                               feature_settings_list=[feature_settings_list[1]])

    # assert lags
    _assert_tsfresh_lags(features_cv=features_cv, lag_settings_list=[lag_settings_list[0]])
    _assert_indicator_lags(features_cv=features_cv, context=context, lag_settings_list=[lag_settings_list[1]])

    # assert counter
    # assert features_cv["Counter"].sum() ==
    print(features_cv["Counter"].sum())



import pandas as pd
import numpy as np

from statsmodels.stats.proportion import multinomial_proportions_confint
import moosir_feature.transformers.feature_cleaning.feature_cleaner as f_cleaner
import moosir_feature.transformers.tsfresh_features.feature_manager as f_tsfresh
import moosir_feature.transformers.basic_features.feature_calculator as f_basic
import moosir_feature.transformers.indicators.feature_manager as f_indicator
import moosir_feature.transformers.feature_selections.feature_selector as f_selector
import moosir_feature.transformers.features_common.calculator as f_common
import moosir_feature.transformers.managers.feature_manager as f_manager


def create_features(instances: pd.DataFrame,
                    feature_creator_manager: f_manager.FeatureCreatorManager
                    ):
    features_all, targets_all, _ = feature_creator_manager.create_features_and_targets(instances=instances)

    features_all = f_common.remove_ohlc_columns_if_exists(data=features_all)

    return dict(features_all=features_all, targets_all=targets_all)


def select_features(features_all: pd.DataFrame,
                    targets_all: pd.DataFrame,
                    feature_selector_manager: f_manager.FeatureSelectorManager,
                    ):
    features, targets, _ = f_common.drop_na(features=features_all, targets=targets_all)

    features_selected, targets_selected = feature_selector_manager.select_features(features=features, targets=targets)

    return dict(features_selected=features_selected, targets_selected=targets_selected)


###############
# feature selection stat significant
##################
def initial_feature_counters(features_all: pd.DataFrame):
    counters = np.zeros(len(features_all.columns))
    feature_keys = features_all.columns
    counter_dic = dict(zip(feature_keys, counters))
    return counter_dic


def update_feature_counters(features_selected: pd.DataFrame, counter_dic: dict):
    for key in features_selected.columns:
        counter_dic[key] += 1
    return counter_dic


class FeatureCVSettings:
    def __init__(self, f_confidence_win_step_perc: float,
                 f_confidence_win_len: int,
                 sample_selection_rand_threshold=1,
                 step_count=-1):
        assert 0 < f_confidence_win_step_perc < 1, "needs to be between 0 and 1"
        assert 0 < sample_selection_rand_threshold <= 1, "needs to be between 0 and 1"
        # assert f_confidence_win_len is int

        self.f_confidence_win_step_perc = f_confidence_win_step_perc
        self.f_confidence_win_len = f_confidence_win_len
        self.sample_selection_rand_threshold = sample_selection_rand_threshold
        self.step_count = step_count


def _remove_small_counter_keys(counter_dics: dict, freq_threshold: int):
    removed = {}
    res = {}
    for key in counter_dics:
        if counter_dics[key] >= freq_threshold:
            res[key] = counter_dics[key]
        else:
            removed[key] = counter_dics[key]

    return res, removed


def calculate_confidence_level(instances: pd.DataFrame,
                               feature_creator_manager: f_manager.FeatureCreatorManager,
                               feature_selector_manager: f_manager.FeatureSelectorManager,
                               feature_cv_settings: FeatureCVSettings):
    step_len = int(feature_cv_settings.f_confidence_win_len * feature_cv_settings.f_confidence_win_step_perc)

    step_count = int(len(instances) / step_len)
    if feature_cv_settings.step_count != -1:
        step_count = feature_cv_settings.step_count
        print(f"step_count overwritten by {step_count}")

    sample_rand_threshold = feature_cv_settings.sample_selection_rand_threshold

    dummy_features = create_features(instances=instances.iloc[:feature_cv_settings.f_confidence_win_len],
                                     feature_creator_manager=feature_creator_manager)

    dummy_features = dummy_features["features_all"]

    counters = initial_feature_counters(features_all=dummy_features)

    # todo: revert it" range(2)
    # for step in range(step_count-1):
    for step in range(step_count - 1):

        # should pick this window?
        rand = np.random.uniform(0, 1)
        if rand > sample_rand_threshold:
            print("sample skipped")
            continue

        start = step * step_len
        end = start + feature_cv_settings.f_confidence_win_len

        if end > len(instances):
            break
        # print('-' * 50)
        # print(f"{start}: {end}")
        active_instances = instances.iloc[start:end]

        all_result = create_features(instances=active_instances,
                                     feature_creator_manager=feature_creator_manager)

        features_all = all_result["features_all"]
        targets_all = all_result["targets_all"]

        s_result = select_features(features_all=features_all,
                                   targets_all=targets_all,
                                   feature_selector_manager=feature_selector_manager
                                   )

        features_selected = s_result["features_selected"]

        counters = update_feature_counters(features_selected=features_selected, counter_dic=counters)

    counters, removed = _remove_small_counter_keys(counter_dics=counters, freq_threshold=1)

    print("removed features for cv: ")
    print(removed)

    # calculate confidence level
    counters_values = np.array(list(counters.values()))

    ## confidence interval
    ci_selected = multinomial_proportions_confint(counters_values, alpha=0.05, method='sison-glaz')

    # prepare output
    features_cdf = pd.DataFrame({"Feature": np.array(list(counters.keys())), "Counter": counters_values})
    features_cvdf = pd.DataFrame(ci_selected, columns=["Min_cv", "Max_cv"])
    features_cvs = features_cdf.join(features_cvdf)

    print('+' * 50)
    print(features_cvs)
    print('+' * 50)
    return features_cvs



import pandas as pd
import numpy as np

from .core import *

FEATURE_PREFIX = "SIG"

FEATURES_SCHEMA_DICS = {"Peaks_All": ["Peaks_All"],
                        "Troughs_All": ["Troughs_All"]
                        }

FEATURES_SETTINGS_DICS = {"Peaks_All": {"extreme_n": 10,
                                        "polyorder": 3,
                                        "smoothing_win_len_perc": 0.20,
                                        "col_name":"SIG-Peaks_All"
                                        },
                          "Troughs_All": {"extreme_n": 10,
                                        "polyorder": 3,
                                        "smoothing_win_len_perc": 0.20,
                                        "col_name":"SIG-Troughs_All"
                                        },
                          }

FEATURES_CALC_DICS = {"Peaks_All": peaks_all,
                       "Troughs_All": troughs_all
                        }

FEATURE_FWD_PREFIX = "Fwd"

def _get_feature_col_name(prefix: str, win_len: int, orig_col_name: str):
    return f"{prefix}{'' if prefix == '' else '-'}{orig_col_name}-{win_len}"


def get_feature_col_full_names(feature_name: str, win_len: int):
    if feature_name in ["Peaks_All", "Troughs_All"]:
        settings = FEATURES_SETTINGS_DICS[feature_name]
        col_pref = settings["col_name"]
        extreme_n = settings["extreme_n"]
        cols = []
        for i in range(extreme_n):
            col = find_extreme_get_col_name(col_name_prefix=col_pref, extreme_ind=i, win_len=win_len)
            cols.append(col)
        return cols

    result = [_get_feature_col_name(prefix=FEATURE_PREFIX, win_len=win_len, orig_col_name=k)
              for k in
              FEATURES_SCHEMA_DICS[feature_name]]
    return result


def get_feature_col_by_subtype_name(feature_name: str, columns: list, is_fwd: bool):
    for col in columns:
        if feature_name in col:
            if is_fwd:
                if FEATURE_FWD_PREFIX in col:
                    return col
            else:
                return col
    raise Exception(f"column not found: {feature_name}")

def get_feature_lag_full_name(feature_name: str, win_len: int, lag_len: int):
    col_names = get_feature_col_full_names(feature_name=feature_name, win_len=win_len)
    results = [f"{col_name}_Lag-{lag_len}" for col_name in col_names]
    return results


def get_feature_fwd_full_name(feature_name: str, win_len: int, fwd_len: int):
    col_names = get_feature_col_full_names(feature_name=feature_name, win_len=win_len)
    results = [f"{col_name}_{FEATURE_FWD_PREFIX}-{fwd_len}" for col_name in col_names]
    return results


def validate_columns(data: pd.DataFrame,
                     feature_names: list,
                     win_lens: list,
                     lag_lens: list = None,
                     forward_lens: list = None):
    assert lag_lens is None or forward_lens is None, "forwards and lags cant be not none at the same time"

    cols = data.columns
    for f in feature_names:
        for w in win_lens:
            if lag_lens:
                for lag in lag_lens:
                    col_names = get_feature_lag_full_name(feature_name=f, win_len=w, lag_len=lag)
                    for col in col_names:
                        assert col in cols, f"missing column {col}"
            elif forward_lens:
                for fwd in forward_lens:
                    col_names = get_feature_fwd_full_name(feature_name=f, win_len=w, fwd_len=fwd)
                    for col in col_names:
                        assert col in cols, f"missing column {col}"
            else:
                col_names = get_feature_col_full_names(feature_name=f, win_len=w)
                for col in col_names:
                    assert col in cols, f"missing column {col}"
    return True


def calc_rolling_feature(ohlc: pd.DataFrame, win_len=10, selected_features=["Peaks_All"],
                         output_col_prefix=FEATURE_PREFIX, verbose=False):
    """
        applies one feature on input data rolling wind
        Note: tsfresh feature result cant be pandas per window
    """
    if verbose:
        print("#######################")
        print("prepare data for signals")
        print("#######################")
    # data = ohlc[["Close"]]
    data = ohlc.copy()
    assert data.index.name == "Timestamp"

    sig_data = data
    if verbose:
        print("#######################")
        print("calculate features")
        print("#######################")


    features_ls = []
    for feature in selected_features:
        arg_vals = {"ohlc": ohlc, "win_len": win_len}
        arg_vals.update(FEATURES_SETTINGS_DICS[feature])
        features_ls.append(FEATURES_CALC_DICS[feature](**arg_vals))

    features_df = pd.concat(features_ls, axis=1)

    assert len(features_df) + win_len == 1 + len(
        data), f"some data missed during calculation. lengths: input {len(data)}, result {len(features_df)}, window: {win_len} "

    return features_df



import pandas as pd
import pytest

from moosir_feature.transformers.tsfresh_features.feature_manager import *


# from src.research.preproc.pipeline_kedro.src.preproc.domain.transformers.tsfresh_features.contexts import *


@pytest.fixture
def ohlc():
    df = pd.DataFrame(data=np.random.rand(100, 4),
                      columns=["Open", "High", "Low", "Close"],
                      index=pd.date_range(start="01/01/2000", periods=100, freq="10T"),
                      )
    df.index.name = "Timestamp"
    return df


"""
todo: test context!!!!
"""


#

def test_feature_manager_calculate_forwards(ohlc: pd.DataFrame):
    # arrange
    win_lens = [80, 90]
    forward_lens = [2, 5]

    feature_names = FEATURES_SCHEMA_DICS.keys()
    # feature_names = ["linear_trend_timewise"]

    # act
    result = calculate_forwards(ohlc=ohlc,
                                win_lens=win_lens,
                                forward_lens=forward_lens,
                                feature_names=feature_names)

    # assert
    print(result)
    print(result.columns)
    assert validate_columns(data=result, feature_names=feature_names, win_lens=win_lens, forward_lens=forward_lens)
    assert len(result) == len(ohlc) - 95, "nan values generated by window still exist"


def test_feature_manager_calculate_features_lags(ohlc: pd.DataFrame):
    # arrange
    win_lens = [80, 90]
    lag_lens = [2, 5]

    feature_names = FEATURES_SCHEMA_DICS.keys()
    # feature_names = ["linear_trend_timewise"]

    # act
    result = calculate_lags(ohlc=ohlc,
                            win_lens=win_lens,
                            lag_lens=lag_lens,
                            feature_names=feature_names)

    # assert
    print(result)
    print(result.columns)
    assert validate_columns(data=result, feature_names=feature_names, win_lens=win_lens, lag_lens=lag_lens)
    assert len(result) == len(ohlc) - 95, "nan values generated by window still exist"


def test_feature_manager_calculate_features_all_indicators(ohlc: pd.DataFrame):
    # arrange
    win_lens = [80, 90]

    feature_names = FEATURES_SCHEMA_DICS.keys()
    # feature_names = ["linear_trend_timewise"]

    # act
    result = calculate_features(ohlc=ohlc,
                                win_lens=win_lens,
                                feature_names=feature_names)

    # assert
    assert validate_columns(data=result, feature_names=feature_names, win_lens=win_lens)
    assert len(result) == len(ohlc) - 90, "nan values generated by window still exist"

#
#
#
#
#
# 
# def test_feature_manager_calculate_features_all_indicators(ohlc: pd.DataFrame):
#     # arrange
#     win_lens = [3, 4]
#
#     feature_names = OPERATORS_IND.keys()
#
#     context = Context()
#
#     # act
#     result = calculate_features(ohlc=ohlc,
#                                 win_lens=win_lens,
#                                 feature_names=feature_names)
#
#     # assert
#     assert context.validate(result, operators=OPERATORS_IND.values(), periods=win_lens)
#     assert len(result) == len(ohlc) - 4, "nan values generated by window still exist"
#
#



from moosir_feature.transformers.indicators.tech_indicators import *
from moosir_feature.transformers.indicators.contexts import OPERATORS_IND
import moosir_feature.transformers.tsfresh_features.feature_calculator  as tsf
import moosir_feature.transformers.signal_features.feature_calculator  as sig


class Settings:
    pass


###############
# features
###############

class FeatureSettings:
    pass


class TsfreshFeatureSettings(FeatureSettings):
    def __init__(self, win_lens, feature_names):
        self.win_lens = win_lens
        if "all" in feature_names:
            self.feature_names = list(tsf.FEATURES_SCHEMA_DICS.keys())
        else:
            self.feature_names = feature_names


class IndicatorFeatureSettings(FeatureSettings):
    def __init__(self, win_lens, feature_names):
        self.win_lens = win_lens  # [80]
        # todo: shameful
        if "all" in feature_names:
            self.feature_names = list(OPERATORS_IND.keys())
        else:
            self.feature_names = feature_names  # [VilliamrOperator.__name__]


class BasicFeatureSettings(FeatureSettings):
    def __init__(self, feature_names):
        self.feature_names = feature_names  # ["Hour"]

class SignalFeatureSettings(FeatureSettings):
    def __init__(self, win_lens, feature_names):
        self.win_lens = win_lens
        if "all" in feature_names:
            self.feature_names = list(sig.FEATURES_SCHEMA_DICS.keys())
        else:
            self.feature_names = feature_names

###############
# lags
###############
class LagSettings(FeatureSettings):
    pass


class TsfreshLagSettings(TsfreshFeatureSettings, LagSettings):
    def __init__(self, feature_parameters, lag_lens):
        super(TsfreshLagSettings, self).__init__(**feature_parameters)
        self.lag_lens = lag_lens


class IndicatorLagSettings(IndicatorFeatureSettings, LagSettings):
    def __init__(self, feature_parameters, lag_lens):
        super(IndicatorLagSettings, self).__init__(**feature_parameters)
        self.lag_lens = lag_lens

class SignalLagSettings(SignalFeatureSettings, LagSettings):
    def __init__(self, feature_parameters, lag_lens):
        super(SignalLagSettings, self).__init__(**feature_parameters)
        self.lag_lens = lag_lens

###############
# targets
###############

class TargetSettings:
    def __init__(self):
        pass


class TsfreshTargetSettings(TargetSettings):
    def __init__(self, win_len,
                 lookahead_len,
                 feature_name,
                 target_col_feature_type,
                 target_col_feature_subtype):
        self.win_len = win_len
        self.lookahead_len = lookahead_len
        self.feature_name = feature_name
        self.target_col_feature_type = target_col_feature_type
        self.target_col_feature_subtype = target_col_feature_subtype


class IndicatorTargetSettings(TargetSettings):
    def __init__(self, win_len,
                 lookahead_len,
                 feature_name,
                 target_col_feature_type,
                 target_col_feature_subtype):
        self.win_len = win_len
        self.lookahead_len = lookahead_len
        self.feature_name = feature_name
        self.target_col_feature_type = target_col_feature_type
        self.target_col_feature_subtype = target_col_feature_subtype

class SignalTargetSettings(TargetSettings):
    def __init__(self, win_len,
                 lookahead_len,
                 feature_name,
                 target_col_feature_type,
                 target_col_feature_subtype):
        self.win_len = win_len
        self.lookahead_len = lookahead_len
        self.feature_name = feature_name
        self.target_col_feature_type = target_col_feature_type
        self.target_col_feature_subtype = target_col_feature_subtype


##############
# feature selector
##############

class FeatureSelectorSettings(TargetSettings):
    def __init__(self, features_to_select_perc):
        self.features_to_select_perc = features_to_select_perc



import pandas as pd
import numpy as np

from sklearn.base import RegressorMixin
from sklearn.model_selection import cross_validate

from .metrics import get_any_metric

class NaiveModel(RegressorMixin):
    """
     - lag target of look_back_len retuns as prediction
    """

    def __init__(self, targets: pd.DataFrame, look_back_len: int):
        # self.rand_seed = rand_seed if rand_seed is not None else np.random.randint(0, 1000)
        self.look_back_len = look_back_len
        self.targets = targets

    def fit(self, X=None, y=None):
        pass

    def predict(self, X):
        vals = self.targets[self.targets.index.isin(X.index)]

        iloc_last = self.targets.index.get_loc(X.index.min())
        val_last = self.targets.iloc[iloc_last - self.look_back_len]
        vals = vals.shift(self.look_back_len).dropna()
        result = np.append(val_last, vals.values.reshape(-1))

        return result

    def get_params(self, deep=False):
        return {'look_back_len': self.look_back_len, 'targets': self.targets}

    def set_params(self, **parameters):
        self.look_back_len = parameters['look_back_len']
        self.targets = parameters['targets']
        return self


class RandomModel(RegressorMixin):
    def __init__(self, min, max):
        # self.rand_seed = rand_seed if rand_seed is not None else np.random.randint(0, 1000)
        self.min = min
        self.max = max

    def fit(self, X=None, y=None):
        pass

    def predict(self, X):
        vals = np.random.uniform(self.min, self.max, (len(X), 1))
        return vals

    def get_params(self, deep=False):
        return {'min': self.min, 'max': self.max}

    def set_params(self, **parameters):
        self.min = parameters['min']
        self.max = parameters['max']
        return self

def run_benchmarking(models: list, cv, features: pd.DataFrame, targets: pd.DataFrame, metrics: list):
    """

    :param models: list of models to be selected from
    :param cv: cv to run the models
    :param features: features
    :param targets: targets
    :return: benchmarking result
      fit_time	score_time test_<metric1> ... test_<metricn>, Model, Split_Cv
      --------	---------- -------------- ... --------------  -----  --------

    """
    cv_results = []
    scorers = get_any_metric(metrics=metrics)
    for model in models:

        cv_result = pd.DataFrame(cross_validate(estimator=model, cv=cv, X=features, y=targets,
                                                scoring=scorers))
        cv_result['Model'] = type(model).__name__
        cv_result['Split_Cv'] = pd.Series(data=cv_result.index.values, index=cv_result.index)

        cv_results.append(cv_result)

    result = pd.concat(cv_results, axis=0).reset_index()
    _ = result.pop("index")
    return result



import pandas as pd

from moosir_feature.transformers.features_cv.feature_cv_manager import FeatureCVSettings, calculate_confidence_level
from moosir_feature.transformers.managers.feature_manager import FeatureCreatorManager, FeatureSelectorManager


class FeatureConfidenceLevelManager:
    def __init__(self, feature_cv_settings: FeatureCVSettings,
                 feature_creator_mgr: FeatureCreatorManager,
                 feature_selector_mgr: FeatureSelectorManager):
        self.feature_cv_settings = feature_cv_settings
        self.feature_creator_mgr = feature_creator_mgr
        self.feature_selector_mgr = feature_selector_mgr

    def calculate_cv(self, instances: pd.DataFrame):
        return calculate_confidence_level(instances=instances,
                                                feature_creator_manager=self.feature_creator_mgr,
                                                feature_selector_manager=self.feature_selector_mgr,
                                                feature_cv_settings=self.feature_cv_settings)


"""
    - create features (indicators) based on
        - https://github.com/PacktPublishing/Machine-Learning-for-Algorithmic-Trading-Second-Edition/blob/master/12_gradient_boosting_machines/04_preparing_the_model_data.ipynb
    - features are CLOSE price centered
    - Todo: some features have period (e.x. 14). why?
    - TODO: returns and some other features are too small due to Forex
"""
import logging

import pandas as pd

from .contexts import Context
from .feature_calculator import *
from .feature_calculator import *
import moosir_feature.transformers.feature_cleaning.feature_cleaner as f_cleaner


def calculate_features(ohlc: pd.DataFrame,
                       win_lens: list,
                       feature_names: list):
    # technical indictors
    ohlc_price, _ = apply_technical_indicators(ohlc_price=ohlc,
                                            look_back_periods=win_lens,
                                            tech_indicator_names=feature_names)

    # remove windows nan related
    max_win_len = max(win_lens)
    ohlc_price = ohlc_price.iloc[max_win_len:]

    # # Historical Returns
    # ohlc_price = apply_historical_return(ohlc_price, periods=win_lens)

    return ohlc_price


def calculate_forwards(ohlc: pd.DataFrame,
                      win_lens: list,
                      forward_periods: list,
                      feature_names: list
                      ):
    # technical indictors
    ohlc_price, tech_indicators = apply_technical_indicators(ohlc_price=ohlc,
                                                             look_back_periods=win_lens,
                                                             tech_indicator_names=feature_names)

    # remove windows nan related
    max_len = max(win_lens)
    ohlc_price = ohlc_price.iloc[max_len:]

    context = Context()

    for indic in tech_indicators:
        col_name_fwds = context.get_column_names(operator=indic,
                                                 period=indic.period,
                                                 operator_name=type(indic).__name__,
                                                 forward_periods=forward_periods)

        col_name_no_lags = context.get_column_names(operator=indic,
                                                    period=indic.period,
                                                    operator_name=type(indic).__name__)
        for f_i in range(len(col_name_no_lags)):
            for fwd_i in range(len(forward_periods)):
                ind = f_i * len(forward_periods)  + fwd_i
                ohlc_price[col_name_fwds[ind]] = ohlc_price[col_name_no_lags[f_i]].shift(-1 * forward_periods[fwd_i])

    # remove windows nan related
    max_len = max(forward_periods)
    ohlc_price = ohlc_price.iloc[:-max_len]


    return ohlc_price


def calculate_lags(ohlc: pd.DataFrame,
                   win_lens: list,
                   lag_lens: list,
                   feature_names: list):


    ohlc_price = ohlc
    operators = []


    ohlc_price, indicators = apply_technical_indicators(ohlc_price=ohlc_price,
                                            look_back_periods=win_lens,
                                            tech_indicator_names=feature_names)
    for ind in indicators:
        ohlc_price = apply_lags(ohlc_price=ohlc_price,
                                lag_periods=lag_lens,
                                ind_periods=win_lens,
                                operator_type=type(ind))



        # for lag_i in range(len(lag_lens)):
        #     for col_i in range(len(col_name_no_lags)):
        # ohlc_price[col_name_lags[lag_i]] = ohlc_price[col_name_no_lags[col_i]].shift(lag_lens[lag_i])

    # remove windows nan related
    max_len = max(lag_lens)
    ohlc_price = ohlc_price.iloc[max_len:]

    return ohlc_price

# todo: wrong
def get_feature_col_by_subtype_name(feature_name: str, columns: list, is_fwd: bool):
    for col in columns:
        if feature_name in col:
            if is_fwd:
                if RESULT_TECH_FWD_PREFIX in col:
                    return col
            else:
                return col
    raise Exception(f"column not found: {feature_name}")






import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal, assert_index_equal, assert_extension_array_equal

from moosir_feature.model_validations.model_validator import *


@pytest.fixture
def ohlc_prices():
    return pd.DataFrame(data=np.random.rand(10, 4), columns=["Open", "High", "Low", "Close"])


def test_shuffle_df():
    # arrange
    sample_n = 10
    feature_n = 3
    feature_cols = [f"f_{i}" for i in range(feature_n)]
    target_col = "target"

    features = pd.DataFrame(data=np.random.rand(sample_n, feature_n),
                            index=pd.date_range("2020/01/01", periods=sample_n),
                            columns=feature_cols)
    targets = pd.DataFrame(data=np.random.rand(sample_n, 1),
                           index=pd.date_range("2020/01/01", periods=sample_n),
                           columns=[target_col]
                           )

    data = pd.concat([features, targets], axis=1)

    # act
    features_shuf, targets_shuf, all_shuf, rand_state = shuffle_df(data=data, target_col=target_col)

    # assert
    assert_frame_equal(features_shuf.sort_index().reset_index(drop=True),
                       features.sort_index().reset_index(drop=True),
                       check_index_type=False)

    assert_frame_equal(targets_shuf.sort_index().reset_index(drop=True),
                       targets.sort_index().reset_index(drop=True),
                       check_index_type=False)

    assert_frame_equal(all_shuf.sort_index().reset_index(drop=True),
                       data.sort_index().reset_index(drop=True),
                       check_index_type=False)


def test_cv_ts_df():
    # arrange
    sample_n = 10
    feature_n = 1

    feature_cols = [f"f_{i}" for i in range(feature_n)]
    target_col = "target"
    features = pd.DataFrame(data=np.arange(0, sample_n * feature_n).reshape(sample_n, -1), columns=feature_cols)
    targets = pd.DataFrame(data=np.arange(0, sample_n * 1).reshape(sample_n, -1), columns=[target_col])

    train_n = 4
    test_n = 2

    # act
    cv = CustomTsCv(train_n=train_n, test_n=test_n, sample_n=sample_n)

    # assert
    splits = list(cv.split(X=features, y=targets))
    print(splits)

    # todo: more asserts


def test_calculate_confidence_interval():
    # arrange
    sample_n = 10
    statistics_n = 2

    statistics_cols = [f"stat_{i}" for i in range(statistics_n)]
    statistics = pd.DataFrame(data=np.arange(0, sample_n * statistics_n).reshape(sample_n, -1),
                              columns=statistics_cols)

    # act
    ci_results = calculate_confidence_interval(statistics=statistics, ignore_small_instances=True)

    # assert
    expected = pd.DataFrame(data={"Lower_2.5": [0.45, 1.45], "Upper_97.5": [17.55, 18.55]}, index=["stat_0", "stat_1"])
    assert_frame_equal(expected, ci_results)




"""
todo:
    - there is way more ways of removing outliers:  Peirce criterion, Grubb's test or Dixon's Q-test, clusters, zscore, ...


"""
import logging
import pandas as pd
from pandas.tseries.frequencies import to_offset
from sklearn.preprocessing import MinMaxScaler


logger = logging.getLogger(__file__)

def clip_quantile(features: pd.DataFrame, lower_quantile=0.01, upper_quantile=0.99):
    return features.clip(features.quantile(lower_quantile), features.quantile(upper_quantile), axis=1)


def _apply_timestamp_ohlc_resample(ohlc_row):

    if len(ohlc_row) == 0:
        # todo: potential unwanted data loss
        # happens normally with none inferable freq (e.x. datatimeindex with missing values like weekends)
        return

    row = ohlc_row.iloc[0].copy()
    row["Open"] = ohlc_row["Open"].iloc[0]
    row["Close"] = ohlc_row["Close"].iloc[-1]
    row["Low"] = ohlc_row["Low"].min()
    row["High"] = ohlc_row["High"].max()

    return row

def timestamp_ohlc_resample(ohlc: pd.DataFrame, resample_freq: str) -> pd.DataFrame:
    """
    - merging ohlc data during resampling
    - note that using first/last on higher frequency is wrong in ohlc context

    """
    assert sorted(["Open", "High", "Low", "Close"]) == sorted(ohlc.columns), "input missing/more ohlc required columns"
    # never use last() cos it hints the model for future
    # never use resample() cos it put nan for weekends and holidays
    requested_freq = to_offset(resample_freq)
    ohlc_freq = ohlc.index.freq

    if ohlc_freq is None:
        logger.warning("input ohlc does not have any frequency, will infer to validate")
        diffs = (ohlc.index[1:] - ohlc.index[:-1])
        ohlc_freq = to_offset(diffs[0])

    if requested_freq <= ohlc_freq:
        logger.warning(f"requested frequency is bigger than input ohlc frequency, no action: {requested_freq } <= {ohlc_freq} ")
        return ohlc

    # ohlc.index = pd.to_datetime(ohlc.index)

    resampled = (ohlc
                 .groupby(pd.Grouper(freq=resample_freq))
                 .apply(_apply_timestamp_ohlc_resample)
                  # for missing data (e.x. weekends) it still creates nan!!
                  # so, rows with all nan likely to be result of this group by - so, delete them all
                 .dropna(how="all")
                 )
    return resampled




def convert_to_pips(data: pd.DataFrame):
    return data.round(4).sub(1).mul(10000)


def apply_basic_cleaning(instances: pd.DataFrame, resample_freq: str):
    # basic transforms
    result = instances.copy()
    result = timestamp_ohlc_resample(ohlc=result, resample_freq=resample_freq)
    result = convert_to_pips(data=result)

    return result


def standardize(data: pd.DataFrame):
    """
    rescale all values to be have mean = 0 and std = 1
    z = (x- mean) / std
    Note:
        - Better not to use it cuz of any normalize assumption to financial data is dangerous (todo: to be confirmed)!!!
        - THE RESULT WILL BE NORMAL DISTRIBUTION
        - standard deviation and mean changed (unlike normalization)
        - z-score standardization
        - good for comparing data with different UoM
        - must for gradient descent based

    """
    raise Exception("Not implemented!!")


def normalize(data: pd.DataFrame):
    """
      rescale all values to be [0,1]
      z = (X - Xmin) / Xmax - Xmin
      Note:
          - DOES NOT change the distribution of the result
          - standard deviation and mean is not changed (unlike standardize)
          - MinMaxScaling is normalization
          - can suppress the effect of outliers (not sure)??
          - must for gradient descent based
          - not required for decision trees
          - you should scale target variable too!!!
      :param target_col: just to make sure that target col is not in this process
      """

    scaler = MinMaxScaler()
    scaler.fit(data)
    transformed = scaler.transform(data)

    transformed_df = pd.DataFrame(data=transformed, index=data.index, columns=data.columns)
    return transformed_df, scaler




import pytest

from moosir_feature.transformers.managers.feature_manager import *
from moosir_feature.transformers.indicators.feature_calculator import *
from moosir_feature.transformers.basic_features.feature_calculator import *


@pytest.fixture
def ohlc():
    df = pd.DataFrame(data=np.random.rand(100, 4),
                      columns=["Open", "High", "Low", "Close"],
                      index=pd.date_range(start="01/01/2000", periods=100, freq="10T"),
                      )
    df.index.name = "Timestamp"
    return df


def _print(features, targets, all):
    print(features)
    print(features.columns)
    print('-' * 50)
    print(targets)
    print(targets.columns)
    print('-' * 50)
    print(all)
    print(all.columns)


def _run_manager_create_feature(ohlc, target_settings, feature_settings_list, lag_settings_list):
    mgr = FeatureCreatorManager(target_settings=target_settings,
                                feature_settings_list=feature_settings_list,
                                lag_settings_list=lag_settings_list
                                )

    # act
    features, clean_data = mgr.create_features(instances=ohlc)

    return features, clean_data


def _run_manager_create_feature_and_target(ohlc, target_settings, feature_settings_list, lag_settings_list):
    mgr = FeatureCreatorManager(target_settings=target_settings,
                                feature_settings_list=feature_settings_list,
                                lag_settings_list=lag_settings_list
                                )

    # act
    features, targets, all = mgr.create_features_and_targets(instances=ohlc)

    return features, targets, all


def test_feature_manager_calculate_feature_when_every_thing_emtpy(ohlc: pd.DataFrame):
    # arrange
    target_settings = TargetSettings()
    feature_settings_list = []
    lag_settings_list = []

    # act
    features, clean_data = _run_manager_create_feature(ohlc=ohlc,
                                                       target_settings=target_settings,
                                                       feature_settings_list=feature_settings_list,
                                                       lag_settings_list=lag_settings_list)

    # _print(features, clean_data, clean_data)

    # assert features
    assert len(features) == 0


def test_feature_manager_calculate_feature_and_target_when_every_thing_emtpy(ohlc: pd.DataFrame):
    # arrange
    target_settings = TargetSettings()
    feature_settings_list = []
    lag_settings_list = []

    # act
    features, targets, all = _run_manager_create_feature_and_target(ohlc=ohlc,
                                                                  target_settings=target_settings,
                                                                  feature_settings_list=feature_settings_list,
                                                                  lag_settings_list=lag_settings_list)

    # _print(features, clean_data, clean_data)

    # assert features
    assert len(features) == 0
    assert len(targets) == 0



import pytest

from moosir_feature.transformers.managers.feature_manager import *
from moosir_feature.transformers.tsfresh_features.feature_calculator import *
from moosir_feature.transformers.indicators.feature_calculator import *
from moosir_feature.transformers.basic_features.feature_calculator import *
import moosir_feature.transformers.signal_features.feature_manager as sig_mgr


@pytest.fixture
def ohlc():
    df = pd.DataFrame(data=np.random.rand(100, 4),
                      columns=["Open", "High", "Low", "Close"],
                      index=pd.date_range(start="01/01/2000", periods=100, freq="10T"),
                      )
    df.index.name = "Timestamp"
    return df


def _print(features, targets, all):
    print(features)
    print(features.columns)
    print('-' * 50)
    print(targets)
    print(targets.columns)
    print('-' * 50)
    print(all)
    print(all.columns)


def _get_target_settings(setting_key: str):
    ts = TsfreshTargetSettings(**{
        'win_len': 80,
        'lookahead_len': 3,
        'feature_name': 'linear_trend_timewise',
        'target_col_feature_type': 'Trend',
        'target_col_feature_subtype': 'Slope',

    })

    ind = IndicatorTargetSettings(**{
        'win_len': 80,
        'lookahead_len': 3,
        'feature_name': RsiOperator.__name__,
        'target_col_feature_type': 'Rsi',
        'target_col_feature_subtype': RsiOperator.get_column_names()[0],
    })

    sig = SignalTargetSettings(**{
        'win_len': 80,
        'lookahead_len': 3,
        'feature_name': "Peaks_All",
        'target_col_feature_type': '',
        'target_col_feature_subtype': 'Peaks_All',
    })

    settings = {"ts": ts, "ind": ind, "sig": sig}
    return settings[setting_key]


def _get_feature_settings(setting_key: str):
    ts = TsfreshFeatureSettings(**{
        'win_lens': [80],
        'feature_names': ["abs_energy"]})

    ind = IndicatorFeatureSettings(**{
        'win_lens': [80],
        'feature_names': [VilliamrOperator.__name__]
    })

    basic = BasicFeatureSettings(**{
        'feature_names': ["Hour"]
    })

    sig = SignalFeatureSettings(**{
        'win_lens': [80],
        'feature_names': ["Peaks_All"]
    })

    settings = {"ts": ts, "ind": ind, "basic": basic, "sig": sig}
    return settings[setting_key]


def _get_lag_settings(setting_key: str):
    ts = TsfreshLagSettings(**{'feature_parameters': {
        'win_lens': [80],
        'feature_names': ["benford_correlation"]

    },
        'lag_lens': [3]
    })

    ind = IndicatorLagSettings(**{'feature_parameters':
        {
            'win_lens': [80],
            'feature_names': [AtrOperator.__name__]
        },
        'lag_lens': [3]
    })

    sig = SignalLagSettings(**{'feature_parameters':
        {
            'win_lens': [80],
            'feature_names': ["Troughs_All"]
        },
        'lag_lens': [3]
    })

    settings = {"ts": ts, "ind": ind, "sig": sig}

    return settings[setting_key]


def _run_manager(ohlc, target_settings, feature_settings_list, lag_settings_list):
    mgr = FeatureCreatorManager(target_settings=target_settings,
                                feature_settings_list=feature_settings_list,
                                lag_settings_list=lag_settings_list
                                )

    # act
    features, targets, all = mgr.create_features_and_targets(instances=ohlc)

    return features, targets, all


def _get_indicator_operators(feature_names: list):
    return [OPERATORS_IND[f] for f in feature_names]


def _assert_tsfresh_features(features, feature_settings_list):
    for f_settings in feature_settings_list:
        assert validate_columns(data=features,
                                feature_names=f_settings.feature_names,
                                win_lens=f_settings.win_lens)


def _assert_signal_features(features, feature_settings_list):
    for f_settings in feature_settings_list:
        assert sig_mgr.validate_columns(data=features,
                                        feature_names=f_settings.feature_names,
                                        win_lens=f_settings.win_lens)


def _assert_tsfresh_lags(features, lag_settings_list):
    for l_settings in lag_settings_list:
        assert validate_columns(data=features,
                                feature_names=l_settings.feature_names,
                                win_lens=l_settings.win_lens)


def _assert_signal_lags(features, lag_settings_list):
    for l_settings in lag_settings_list:
        assert sig_mgr.validate_columns(data=features,
                                        feature_names=l_settings.feature_names,
                                        win_lens=l_settings.win_lens)


def _assert_indicator_lags(features, context, lag_settings_list):
    for l_settings in lag_settings_list:
        lag_ops = _get_indicator_operators(l_settings.feature_names)
        assert context.validate(features,
                                operators=lag_ops,
                                periods=l_settings.win_lens,
                                lag_periods=l_settings.lag_lens)


def _assert_indicator_features(context, features, feature_settings_list):
    for f_settings in feature_settings_list:
        feature_ops = _get_indicator_operators(f_settings.feature_names)
        assert context.validate(features, operators=feature_ops, periods=f_settings.win_lens)


# 
def test_feature_manager_calculate_basic_features(ohlc: pd.DataFrame):
    # arrange
    target_settings = _get_target_settings("ind")
    feature_settings = _get_feature_settings("basic")
    lag_settings_list = [_get_lag_settings("ind")]

    # act
    features, targets, all = _run_manager(ohlc=ohlc,
                                          target_settings=target_settings,
                                          feature_settings_list=[feature_settings],
                                          lag_settings_list=lag_settings_list)

    _print(features, targets, all)
    # assert
    context = Context()

    # assert features
    assert validate_basic_feature_col(data=features, feature_names=feature_settings.feature_names)

    # assert targets
    fwd_ops = _get_indicator_operators([target_settings.feature_name])
    assert context.validate(targets,
                            operators=fwd_ops,
                            periods=[target_settings.win_len],
                            forward_periods=[target_settings.lookahead_len])

    # assert lags
    _assert_indicator_lags(features=features, context=context, lag_settings_list=lag_settings_list)


def test_feature_manager_calculate_multiple_settings(ohlc: pd.DataFrame):
    # arrange
    target_settings = _get_target_settings("ts")
    feature_settings_list = [_get_feature_settings("ts"), _get_feature_settings("ind")]
    lag_settings_list = [_get_lag_settings("ts"), _get_lag_settings("ind")]

    # act
    features, targets, all = _run_manager(ohlc=ohlc,
                                          target_settings=target_settings,
                                          feature_settings_list=feature_settings_list,
                                          lag_settings_list=lag_settings_list)

    # assert
    # assert features
    context = Context()
    _assert_tsfresh_features(features=features, feature_settings_list=[feature_settings_list[0]])
    _assert_indicator_features(features=features, context=context, feature_settings_list=[feature_settings_list[1]])

    # assert targets
    # todo: cant validate subtype!!!
    # assert validate_columns(data=targets,
    #                         feature_names=[target_settings.feature_name],
    #                         win_lens=[target_settings.win_len])

    # assert lags
    _assert_tsfresh_lags(features=features, lag_settings_list=[lag_settings_list[0]])
    _assert_indicator_lags(features=features, context=context, lag_settings_list=[lag_settings_list[1]])


def test_feature_manager_calculate_tsfresh(ohlc: pd.DataFrame):
    # arrange
    target_settings = _get_target_settings("ts")
    feature_settings_list = [_get_feature_settings("ts")]
    lag_settings_list = [_get_lag_settings("ts")]

    # act
    features, targets, all = _run_manager(ohlc=ohlc,
                                          target_settings=target_settings,
                                          feature_settings_list=feature_settings_list,
                                          lag_settings_list=lag_settings_list)

    # assert
    # assert features
    _assert_tsfresh_features(features=features, feature_settings_list=feature_settings_list)

    # assert targets
    # todo: cant validate subtype!!!
    # assert validate_columns(data=targets,
    #                         feature_names=[target_settings.feature_name],
    #                         win_lens=[target_settings.win_len])

    # assert lags
    _assert_tsfresh_lags(features=features, lag_settings_list=lag_settings_list)


def test_feature_manager_calculate_indicators_tsfresh_mix(ohlc: pd.DataFrame):
    # arrange
    target_settings = _get_target_settings("ind")
    feature_settings_list = [_get_feature_settings("ts")]
    lag_settings_list = [_get_lag_settings("ind")]

    # act
    features, targets, all = _run_manager(ohlc=ohlc,
                                          target_settings=target_settings,
                                          feature_settings_list=feature_settings_list,
                                          lag_settings_list=lag_settings_list)

    # assert
    context = Context()

    # assert len(targets.columns) == 1

    # assert features
    _assert_tsfresh_features(features=features, feature_settings_list=feature_settings_list)

    # assert targets
    fwd_ops = _get_indicator_operators([target_settings.feature_name])
    assert context.validate(targets,
                            operators=fwd_ops,
                            periods=[target_settings.win_len],
                            forward_periods=[target_settings.lookahead_len])

    # assert lags
    _assert_indicator_lags(features=features, context=context, lag_settings_list=lag_settings_list)


# 
def test_feature_manager_calculate_indicators(ohlc: pd.DataFrame):
    # arrange
    target_settings = _get_target_settings("ind")
    feature_settings_list = [_get_feature_settings("ind")]
    lag_settings_list = [_get_lag_settings("ind")]

    # act
    features, targets, all = _run_manager(ohlc=ohlc,
                                          target_settings=target_settings,
                                          feature_settings_list=feature_settings_list,
                                          lag_settings_list=lag_settings_list)

    # assert
    context = Context()

    # assert features
    _assert_indicator_features(features=features, context=context, feature_settings_list=feature_settings_list)

    # assert targets
    fwd_ops = _get_indicator_operators([target_settings.feature_name])
    assert context.validate(targets,
                            operators=fwd_ops,
                            periods=[target_settings.win_len],
                            forward_periods=[target_settings.lookahead_len])

    # assert lags
    _assert_indicator_lags(features=features, context=context, lag_settings_list=lag_settings_list)


def test_feature_manager_calculate_signals(ohlc: pd.DataFrame):
    # arrange
    target_settings = _get_target_settings("sig")
    feature_settings_list = [_get_feature_settings("sig")]
    lag_settings_list = [_get_lag_settings("sig")]

    # act
    features, targets, all = _run_manager(ohlc=ohlc,
                                          target_settings=target_settings,
                                          feature_settings_list=feature_settings_list,
                                          lag_settings_list=lag_settings_list)

    _print(features, targets, all)

    # assert
    _assert_signal_features(features=features, feature_settings_list=feature_settings_list)
    _assert_signal_lags(features=features, lag_settings_list=lag_settings_list)




import pytest

from moosir_feature.transformers.managers.feature_manager import *
from moosir_feature.transformers.tsfresh_features.feature_calculator import *
from moosir_feature.transformers.indicators.feature_calculator import *
from moosir_feature.transformers.basic_features.feature_calculator import *


@pytest.fixture
def ohlc():
    df = pd.DataFrame(data=np.random.rand(100, 4),
                      columns=["Open", "High", "Low", "Close"],
                      index=pd.date_range(start="01/01/2000", periods=100, freq="10T"),
                      )
    df.index.name = "Timestamp"
    return df


def _print(features, targets, all):
    print(features)
    print(features.columns)
    print('-' * 50)
    print(targets)
    print(targets.columns)
    print('-' * 50)
    print(all)
    print(all.columns)


def _get_target_settings(setting_key: str):
    returns = IndicatorTargetSettings(**{
        'win_len': 80,
        'lookahead_len': 3,
        'feature_name': ReturnsOperator.__name__,
        'target_col_feature_type': 'Return',
        'target_col_feature_subtype': ReturnsOperator.get_column_names()[0],
    })

    max_returns = IndicatorTargetSettings(**{
        'win_len': 80,
        'lookahead_len': 3,
        'feature_name': MaxPriceAndReturn.__name__,
        'target_col_feature_type': 'Return',
        'target_col_feature_subtype': MaxPriceAndReturn.get_column_names()[1],
    })

    var_returns = IndicatorTargetSettings(**{
        'win_len': 80,
        'lookahead_len': 3,
        'feature_name': ReturnVarianceOperator.__name__,
        'target_col_feature_type': 'Return-Var',
        'target_col_feature_subtype': ReturnVarianceOperator.get_column_names()[0],
    })

    settings = {"returns": returns, "max_returns": max_returns, "var_returns": var_returns}
    return settings[setting_key]


def _get_feature_settings(setting_key: str):
    ind = IndicatorFeatureSettings(**{
        'win_lens': [70],
        'feature_names': [ReturnsOperator.__name__,
                          MaxPriceAndReturn.__name__,
                          MinPriceAndReturn.__name__,
                          ReturnVarianceOperator.__name__]
    })

    settings = {"ind": ind}
    return settings[setting_key]


def _get_lag_settings(setting_key: str):
    ind = IndicatorLagSettings(**{'feature_parameters':
        {
            'win_lens': [30],
            'feature_names': [ReturnsOperator.__name__,
                              MaxPriceAndReturn.__name__,
                              MinPriceAndReturn.__name__,
                              ReturnVarianceOperator.__name__]
        },
        'lag_lens': [3]
    })
    settings = {"ind": ind}
    return settings[setting_key]


def _run_manager(ohlc, target_settings, feature_settings_list, lag_settings_list):
    mgr = FeatureCreatorManager(target_settings=target_settings,
                                feature_settings_list=feature_settings_list,
                                lag_settings_list=lag_settings_list
                                )

    # act
    features, targets, all = mgr.create_features_and_targets(instances=ohlc)

    return features, targets, all


def _get_indicator_operators(feature_names: list):
    return [OPERATORS_IND[f] for f in feature_names]


def _assert_tsfresh_features(features, feature_settings_list):
    for f_settings in feature_settings_list:
        assert validate_columns(data=features,
                                feature_names=f_settings.feature_names,
                                win_lens=f_settings.win_lens)


def _assert_tsfresh_lags(features, lag_settings_list):
    for l_settings in lag_settings_list:
        assert validate_columns(data=features,
                                feature_names=l_settings.feature_names,
                                win_lens=l_settings.win_lens)


def _assert_indicator_lags(features, context, lag_settings_list):
    for l_settings in lag_settings_list:
        lag_ops = _get_indicator_operators(l_settings.feature_names)
        assert context.validate(features,
                                operators=lag_ops,
                                periods=l_settings.win_lens,
                                lag_periods=l_settings.lag_lens)


def _assert_indicator_features(context, features, feature_settings_list):
    for f_settings in feature_settings_list:
        feature_ops = _get_indicator_operators(f_settings.feature_names)
        assert context.validate(features, operators=feature_ops, periods=f_settings.win_lens)


# 
def test_feature_manager_calculate_returns_features(ohlc: pd.DataFrame):
    # arrange
    target_settings = _get_target_settings("returns")
    feature_settings_list = [_get_feature_settings("ind")]
    lag_settings_list = [_get_lag_settings("ind")]

    # act
    features, targets, all = _run_manager(ohlc=ohlc,
                                          target_settings=target_settings,
                                          feature_settings_list=feature_settings_list,
                                          lag_settings_list=lag_settings_list)

    _print(features, targets, all)
    # assert
    context = Context()

    # assert features
    _assert_indicator_features(features=features, context=context, feature_settings_list=feature_settings_list)

    # assert targets
    fwd_ops = _get_indicator_operators([target_settings.feature_name])
    assert context.validate(targets,
                            operators=fwd_ops,
                            periods=[target_settings.win_len],
                            forward_periods=[target_settings.lookahead_len])

    # assert lags
    _assert_indicator_lags(features=features, context=context, lag_settings_list=lag_settings_list)


@pytest.mark.skip("not sure how to validate when target ind has multiple columns")
def test_feature_manager_calculate_max_returns_targets(ohlc: pd.DataFrame):
    # arrange
    target_settings = _get_target_settings("max_returns")
    feature_settings_list = [_get_feature_settings("ind")]
    lag_settings_list = [_get_lag_settings("ind")]

    # act
    features, targets, all = _run_manager(ohlc=ohlc,
                                          target_settings=target_settings,
                                          feature_settings_list=feature_settings_list,
                                          lag_settings_list=lag_settings_list)

    _print(features, targets, all)
    # assert
    context = Context()

    # assert features
    _assert_indicator_features(features=features, context=context, feature_settings_list=feature_settings_list)

    # assert targets
    fwd_ops = _get_indicator_operators([target_settings.feature_name])
    assert context.validate(targets,
                            operators=fwd_ops,
                            periods=[target_settings.win_len],
                            forward_periods=[target_settings.lookahead_len])

    # assert lags
    _assert_indicator_lags(features=features, context=context, lag_settings_list=lag_settings_list)

def test_feature_manager_calculate_var_returns_features(ohlc: pd.DataFrame):
    # arrange
    target_settings = _get_target_settings("var_returns")
    feature_settings_list = [_get_feature_settings("ind")]
    lag_settings_list = [_get_lag_settings("ind")]

    # act
    features, targets, all = _run_manager(ohlc=ohlc,
                                          target_settings=target_settings,
                                          feature_settings_list=feature_settings_list,
                                          lag_settings_list=lag_settings_list)

    _print(features, targets, all)
    # assert
    context = Context()

    # assert features
    _assert_indicator_features(features=features, context=context, feature_settings_list=feature_settings_list)

    # assert targets
    fwd_ops = _get_indicator_operators([target_settings.feature_name])
    assert context.validate(targets,
                            operators=fwd_ops,
                            periods=[target_settings.win_len],
                            forward_periods=[target_settings.lookahead_len])

    # assert lags
    _assert_indicator_lags(features=features, context=context, lag_settings_list=lag_settings_list)


"""
this is to clean, extract features, ... from news dataprovider (so far just investpy or investing.com)
"""
import pandas as pd
import pytz
import calendar

def _get_months_abrs():
    month_abrs = list(calendar.month_abbr)
    month_abrs.remove('')
    return month_abrs


def _clean_events_col(events_df: pd.DataFrame):
    assert "event" in events_df.columns, "input needs to have event column"

    # noisy word
    month_abrs = _get_months_abrs()
    other_noise_strings = ["MoM", "YoY", "QoQ", "Q1", "Q2", "Q3", "Q4"]

    noise_strings = month_abrs + other_noise_strings

    for noise_string in noise_strings:
        events_df["event"] = events_df["event"].apply(lambda x: x.replace(f"({noise_string})", "").strip())

    # to lowercase, no space
    events_df["event"] = events_df["event"].replace({' ': '_', '-': '_'}, regex=True).str.lower()

    # make event currency related (e.x. usd_cpi and eur_cpi)
    # clean_data["event"] = clean_data["currency"].str.lower() + "_" + clean_data["event"]

    return events_df


# todo: why all day and tentative is mapped like this?
def _clean_time_col(t: str):
    replace_val = "00:00"
    noise_strings = ["", "All Day", "Tentative"]

    for noise_string in noise_strings:
        if t == noise_string:
            return replace_val
    return t


# todo: get just actuals with correct values, it drops na (none measureable events will drop)

def _clean_impact_column(data: pd.DataFrame, col_name: str):
    column_clean = data[col_name].fillna(pd.np.nan
                                         ).dropna(
    ).replace({'K': '*1e3', 'M': '*1e6', 'B': '*1e9', '%': '/100', ',': ''}, regex=True
              ).map(pd.eval).to_frame()

    data[col_name] = column_clean[col_name]
    data = data.dropna()
    return data


def clean_data(news_data: pd.DataFrame,
               currencies=["USD", "EUR"],
               datetime_format='%d/%m/%Y %H:%M',
               data_provider_timezone="GMT"):
    """
    * todo:
        * some news has time of "All Day". here we just put them as the begining of the data i.e. convert to 0.00
        * when cleaning forecast and actual, none measureable events will drop, i.e. if they have None value we just dropped them, they can be important tho
    """

    expected_cols = ['date', 'time', 'zone', 'currency', 'importance', 'event', 'actual', 'forecast', 'previous']
    assert set(news_data.columns) == set(
        expected_cols), f"expected columns are missing: expected: {expected_cols}, actual: {news_data.columns}"

    data = news_data.copy()

    # timezone
    data_provider_tz = pytz.timezone(data_provider_timezone)
    print(f"data provider timezone: {data_provider_tz}")

    # make timestamps
    data["time"] = data["time"].apply(lambda t: _clean_time_col(t))
    data["Timestamp"] = pd.to_datetime(data["date"] + " " + data["time"], format=datetime_format)
    data.pop("time")
    data.pop("date")

    # timezone
    data = data.set_index("Timestamp").tz_localize(data_provider_tz).tz_convert(pytz.utc).tz_localize(tz=None)
    data = data.reset_index()

    # filter currencies
    data = data[data["currency"].isin(currencies)]

    data = _clean_events_col(data)

    data = _clean_impact_column(data=data, col_name="actual")
    data = _clean_impact_column(data=data, col_name="forecast")
    data = _clean_impact_column(data=data, col_name="previous")

    return data

# DATA_PROVIDER = "investingcom"
# dal = DataAccessLayer(data_provider=DATA_PROVIDER, resolution="random", symbol="random")
# news_data = dal.get_news_data_all(if_raw=True)
#
# clean_data(news_data=news_data)



import pytest
import numpy as np
import pandas as pd

from moosir_feature.transformers.feature_selections.feature_selector import *


SAMPLE_N = 100
FEATURE_N = 10

@pytest.fixture
def features():
    return pd.DataFrame(data=np.random.rand(SAMPLE_N, FEATURE_N), columns=[ f"Feature_{i}" for i in range(FEATURE_N)])

@pytest.fixture
def targets():
    return pd.DataFrame(data=np.random.rand(SAMPLE_N, 1), columns=["Target"])


def test_remove_features_recursively(features, targets):
    # arrange
    n_features_to_select = 2

    # act
    result = remove_features_recursively(features=features, targets=targets, n_features_to_select=n_features_to_select)

    # assert
    print(result)
    assert len(result.columns) == n_features_to_select



import pandas as pd
import numpy as np
import scipy.signal as signal

def _find_extreme_inds(data: np.ndarray, comparer=np.greater):
    return signal.argrelextrema(data, comparer)

def troughs_all(ohlc: pd.DataFrame,
                win_len: int,
                extreme_n: int,
                polyorder: int,
                smoothing_win_len_perc: float,
                col_name: str
                ):
    """
    - find n number of troughs in the window (wind len) after smoothing
    - returns distance to last low price in the window (not-normalized by current price)
    """
    assert "Low" in ohlc.columns, "column Low is missing"
    data = ohlc[["Low"]].copy()
    troughs = find_extreme(data=data,
                           win_len=win_len,
                           extreme_n=extreme_n,
                           polyorder=polyorder,
                           smoothing_win_len_perc=smoothing_win_len_perc,
                           comparer=np.less,
                           col_name=col_name)

    return troughs


def peaks_all(ohlc: pd.DataFrame,
              win_len: int,
              extreme_n: int,
              polyorder: int,
              smoothing_win_len_perc: float,
              col_name: str
              ):
    """
        - find n number of peaks in the window (wind len) after smoothing
        - returns distance to last high price in the window (not normalized by current price)
    """

    assert "High" in ohlc.columns, "column High is missing"
    data = ohlc[["High"]].copy()
    peaks = find_extreme(data=data,
                         win_len=win_len,
                         extreme_n=extreme_n,
                         polyorder=polyorder,
                         smoothing_win_len_perc=smoothing_win_len_perc,
                         comparer=np.greater,
                         col_name=col_name)

    return peaks

def find_extreme_get_col_name(col_name_prefix, extreme_ind, win_len):
    return f"{col_name_prefix}_{extreme_ind}-{win_len}"


def find_extreme(data: pd.DataFrame,
                 win_len: int,
                 extreme_n: int,
                 polyorder: int,
                 smoothing_win_len_perc: float,
                 comparer,
                 col_name: str):

    def _find_extreme(win: pd.DataFrame,
                      extreme_n: int,
                      polyorder: int,
                      smoothing_win_len_perc: float,
                      comparer):
        # return 1
        df = data.loc[win.index]
        win_data = df["_data"]

        smooth_win_len = int(len(win_data) * smoothing_win_len_perc)
        if smooth_win_len % 2 == 0:
            smooth_win_len = smooth_win_len + 1
        if smooth_win_len < polyorder:
            polyorder = smooth_win_len - 1

        smooth_d = signal.savgol_filter(win_data, window_length=smooth_win_len, polyorder=polyorder)
        inds = _find_extreme_inds(data=smooth_d, comparer=comparer)
        peaks = win_data.iloc[inds].values.reshape(-1)
        peaks = peaks - win_data.iloc[-1]

        last_row = data.loc[win.index].iloc[-1]
        for i in range(extreme_n):
            if i < len(peaks):
                # todo: shame
                col = find_extreme_get_col_name(col_name_prefix=col_name, extreme_ind=i, win_len=win_len)
                last_row[col]=peaks[i]

        result.append(last_row)

        return 1

    data.columns = ["_data"]

    for i in range(extreme_n):
        col = find_extreme_get_col_name(col_name_prefix=col_name, extreme_ind=i, win_len=win_len)
        data[col] = np.nan

    result = []
    _ = data.rolling(window=win_len).apply(lambda x: _find_extreme(win=x,
                                                                   extreme_n=extreme_n,
                                                                   polyorder=polyorder,
                                                                   smoothing_win_len_perc=smoothing_win_len_perc,
                                                                   comparer=comparer)
                                       , raw=False)
    result_df = pd.DataFrame(result)

    _ = result_df.pop("_data")

    return result_df


from .contexts import *
from .tech_indicators import *
from typing import Dict
from .constants import *




def create_operator(period: int,
                     prefix: str,
                     indic: str):
    klass = globals()[indic]
    instance = klass(prefix=prefix, period=period)
    return instance

def create_operators(look_back_periods: list,
                     prefix: str,
                     tech_indicator_names: list):
    indicators = []
    for indic in tech_indicator_names:
        for period in look_back_periods:
            instance = create_operator(period=period, prefix=prefix, indic=indic)
            indicators.append(instance)

    return indicators


def apply_technical_indicators(ohlc_price,
                               look_back_periods: list,
                               tech_indicator_names: list) -> pd.DataFrame:
    """
    Parameters
    ----------
    ohlc_price: at least ohlc columns
    look_back_periods: all indicators will apply on every look back period
    tech_indicator_names: list of indicator names

    Returns
    -------
    ohlc price + indicators for every period
    format of columns: {constants.RESULT_TECH_IND_PREFIX}{look_back_period:02}_{indicator name}"
    """
    tech_indicators = create_operators(look_back_periods=look_back_periods,
                                       tech_indicator_names=tech_indicator_names,
                                       prefix=RESULT_TECH_IND_PREFIX)

    ohlc_price = run_operators(ohlc_price, tech_indicators)

    return ohlc_price, tech_indicators


def run_operators(ohlc_price, operators):
    for indic in operators:
        ohlc_price = indic.apply(data=ohlc_price)
    return ohlc_price


def remove_ohlc(ohlc_price):
    _ = ohlc_price.pop("Open")
    _ = ohlc_price.pop("High")
    _ = ohlc_price.pop("Low")
    _ = ohlc_price.pop("Close")
    return ohlc_price

# todo: not needed cos moved to indicators, but because of HighestReturn ind it is needed
def apply_max_return_and_price(ohlc_price, periods):
    logger = logging.getLogger(__name__)
    logger.info("Forward max price =================================")
    operators = []
    for period in periods:
        operators.append(MaxPriceAndReturn(prefix=RESULT_TECH_IND_PREFIX,
                                           period=period))

    ohlc_price = run_operators(ohlc_price=ohlc_price, operators=operators)
    return ohlc_price


# todo: not needed cos moved to indicators, but because of HighestReturn ind it is needed
def apply_min_price_and_return(ohlc_price, periods):
    logger = logging.getLogger(__name__)
    logger.info("Forward max price =================================")
    operators = []
    for period in periods:
        operators.append(MinPriceAndReturn(prefix=RESULT_TECH_IND_PREFIX,
                                           period=period))

    ohlc_price = run_operators(ohlc_price=ohlc_price, operators=operators)
    return ohlc_price


def apply_forward_highest_return(ohlc_price, periods):
    logger = logging.getLogger(__name__)
    """
        - todo: so slow cos of if-else
        - returns highest value return with sign (can be from min (neg) or max (pos)
    """
    logger.info("Forward highest return =================================")
    operators = []

    context = Context()

    for period in periods:
        max_cols = context.get_column_names(operator=MaxPriceAndReturn, period=period)[1]
        min_cols = context.get_column_names(operator=MinPriceAndReturn, period=period)[1]

        operators.append(ForwardHighestReturn(
            prefix=RESULT_TECH_FWD_PREFIX,
            prefix_max_return=max_cols,
            prefix_min_return=min_cols,
            period=period))

    ohlc_price = run_operators(ohlc_price=ohlc_price, operators=operators)
    return ohlc_price



def apply_lags(ohlc_price,
               lag_periods,
               ind_periods: list,
               operator_type=ReturnsOperator):
    logger = logging.getLogger(__name__)
    """
        - todo: so slow cos of if-else
        - returns highest value return with sign (can be from min (neg) or max (pos)
    """
    logger.info("Forward highest return =================================")
    operators = []

    context = Context()

    for period in lag_periods:
        for ind_period in ind_periods:
            return_cols = context.get_column_names(operator=operator_type,
                                                   period=ind_period)
            for return_col in return_cols:
                operators.append(Lag(
                    prefix=RESULT_TECH_LAG_PREFIX,
                    prefix_operator=return_col,
                    period=period))

        ohlc_price = run_operators(ohlc_price=ohlc_price, operators=operators)
    return ohlc_price


def apply_candle_stick(ohlc_price):
    logger = logging.getLogger(__name__)
    logger.info("candle sticks =================================")

    operators = [CandleStickOperators(prefix=RESULT_TECH_IND_PREFIX)]

    ohlc_price = run_operators(ohlc_price=ohlc_price, operators=operators)

    return ohlc_price


# def apply_forward_amplitude_labeler(ohlc_price, periods):
#     # todo: period has no meaning in amplitude!!!
#
#     logger = logging.getLogger(__name__)
#     """
#         - based on tseries_patterns github repo and AmplitudeBasedLabeler
#         - based on trend assign +1, 0, -1
#     """
#     logger.info("Forward highest return =================================")
#     operators = []
#
#     for period in periods:
#         operators.append(AmplitudeBasedLabelerOperator(
#             prefix=RESULT_TECH_FWD_PREFIX,
#             period=period))
#
#     ohlc_price = run_operators(ohlc_price=ohlc_price, operators=operators)
#     return ohlc_price


"""
Time features:
    - calculating time related stuff e.x. time, hour, london or NY sessions minutes to, ...
    - todo:
        - holidays
        - more meaningful measures (it s just time difference in minute, not a proper seasonality for market times
          or higher resolutions of dayofweek, ...)
"""

import pandas as pd
from datetime import datetime, time

# NewYork market open/close hours
NY_OPEN_TIME_EET = time(hour=16, minute=30)
NY_CLOSE_TIME_EET = time(hour=23, minute=00)

# London market open/close hours
LDN_OPEN_TIME_EET = time(hour=10, minute=00)
LDN_CLOSE_TIME_EET = time(hour=18, minute=30)

UNIVERSAL_TIMEZONE = "EET"


class FeatureCalculationSettings:
    def TimeFeatureSettings(self):
        return ["Hour", "Ampm", "Dow", "Market-Open-Close-Dist"]


def _stupid_python_time_delta_min(d1: datetime, d2: datetime):
    if d1 > d2:
        return int((d1 - d2).seconds / 60)
    else:
        return int((d2 - d1).seconds / 60 * -1)


def _calculate_time_difference_hour(data: pd.DataFrame, selected_time: time, out_col_name: str):
    # open_time = datetime.strptime(f"FEB 2 2022 {selected_time}", '%b %d %Y %I:%M%p')

    open_time = selected_time
    data["__temp"] = data.index

    # data["__temp"].apply(lambda x: x.replace(hour=0))

    # data["__temp2"] = data["__temp"].replace(hour=13)
    data[f"{out_col_name}"] = data["__temp"].apply(
        lambda x: _stupid_python_time_delta_min(x, datetime.combine(x.date(), open_time)))

    _ = data.pop("__temp")


def calculate_features(ohlc: pd.DataFrame, feature_names: list) -> pd.DataFrame:
    for feature in feature_names:
        if feature == "Hour":
            ohlc["Hour"] = ohlc.index.hour
        elif feature == "Ampm":
            ohlc["__temp"] = ohlc.index.hour
            ohlc["Ampm"] = ohlc["__temp"].apply(lambda x: 1 if x > 12 else -1)
            _ = ohlc.pop("__temp")
        elif feature == "Dow":
            ohlc["Dow"] = ohlc.index.tz_localize(UNIVERSAL_TIMEZONE).day_of_week
        elif feature == "Market-Open-Close-Dist":
            _calculate_time_difference_hour(data=ohlc, selected_time=NY_OPEN_TIME_EET, out_col_name="Ny-Open-Min")
            _calculate_time_difference_hour(data=ohlc, selected_time=NY_CLOSE_TIME_EET, out_col_name="Ny-Close-Min")

            _calculate_time_difference_hour(data=ohlc, selected_time=LDN_OPEN_TIME_EET, out_col_name="Ldn-Open-Min")
            _calculate_time_difference_hour(data=ohlc, selected_time=LDN_CLOSE_TIME_EET, out_col_name="Ldn-Close-Min")

        else:
            raise Exception(f"feature is not implemented: {feature}")

    return ohlc


def validate_basic_feature_col(data: pd.DataFrame, feature_names: list):
    for fn in feature_names:
        if fn not in data.columns:
            return False
    return True


if __name__ == '__main__':
    DATA_PATH = r'C:\datalake-train\moosir\preproc\data\01_raw\close_market_data.csv'
    data = pd.read_csv(DATA_PATH, index_col="Timestamp", parse_dates=True)

    data = data.iloc[2000:2030]
    res = calculate_features(ohlc=data, feature_names=["Market-Open-Close-Dist"])
    print(res)



import pandas as pd
import numpy as np
import re

class IDataAccessLayer:
    def save_news_data(self, news_df: pd.DataFrame, month, year, if_raw):
        pass

    def get_news_data(self, month: int, year: int, if_raw: bool, if_format: bool):
        pass

    def get_news_data_all(self, if_raw):
        pass


class H5DataAccessLayer(IDataAccessLayer):
    def __init__(self, file_path, data_provider):
        self.file_path = file_path
        self.data_provider = data_provider

    def save_news_data(self, news_df: pd.DataFrame, month, year, if_raw):
        key = self._get_news_h5_key(if_raw=if_raw, year=year, month=month)
        news_df.to_hdf(self.file_path, key)

    def get_news_data(self, month: int, year: int, if_raw: bool, if_format: bool):
        key = self._get_news_h5_key(month=month, year=year, if_raw=if_raw)
        raw = pd.read_hdf(self.file_path, key)
        return self._format_raw_data(raw_data=raw)

    def get_news_data_all(self, if_raw):
        store = pd.HDFStore(self.file_path)
        keys = store.keys()
        store.close()

        pattern_str = f"^/{self.data_provider}-(.*)-raw$" if if_raw else f"^/{self.data_provider}-(.*)$"
        pattern = re.compile(pattern_str)

        dfs = []
        for key in keys:
            pattern_result = pattern.search(key)

            if pattern_result:
                print(f"reading key: {key}")
                dfs.append(pd.read_hdf(self.file_path, key))

        result = pd.concat(dfs, axis=0) if len(dfs) > 0 else None

        result = self._format_raw_data(raw_data=result)

        return result

    def _format_raw_data(self, raw_data: pd.DataFrame):
        return raw_data.set_index("id").sort_index().drop_duplicates()

    def _get_news_h5_key(self, month, year, if_raw):
        key = self.data_provider if not if_raw else f"{self.data_provider}-{year}-{month}-raw"
        return key



import pandas as pd

from .feature_calculator import *


def calculate_features(ohlc: pd.DataFrame, win_lens: list, feature_names: list):
    results = []
    for win_len in win_lens:
        result = calc_rolling_feature(ohlc=ohlc, win_len=win_len, selected_features=feature_names)
        results.append(result)

    result = pd.concat(results, axis=1, join='inner')

    return result


def calculate_lags(ohlc:pd.DataFrame,
                    win_lens:int,
                    lag_lens:list,
                    feature_names:list):

    features = calculate_features(ohlc=ohlc, win_lens=win_lens, feature_names=feature_names)

    for f_name in feature_names:
        for lag_len in lag_lens:
            for win in win_lens:
                full_name_lags = get_feature_lag_full_name(feature_name=f_name, win_len=win, lag_len=lag_len)
                full_name_cols = get_feature_col_full_names(feature_name=f_name, win_len=win)
                for i in range(len(full_name_lags)):
                    features[full_name_lags[i]] = features[full_name_cols[i]].shift(lag_len)

    features = features.iloc[max(lag_lens):]

    return features


def calculate_forwards(ohlc: pd.DataFrame,
                      win_lens: list,
                      forward_lens: list,
                      feature_names: list
                      ):
    features = calculate_features(ohlc=ohlc, win_lens=win_lens, feature_names=feature_names)

    for f_name in feature_names:
        for fwd_len in forward_lens:
            for win in win_lens:
                full_name_fwds = get_feature_fwd_full_name(feature_name=f_name, win_len=win, fwd_len=fwd_len)
                full_name_cols = get_feature_col_full_names(feature_name=f_name, win_len=win)
                for i in range(len(full_name_fwds)):
                    features[full_name_fwds[i]] = features[full_name_cols[i]].shift(-1*fwd_len)


    features = features.iloc[:-1*max(forward_lens)]

    return features






import logging

import talib
from talib import *
import pandas as pd
import numpy as np

DECIMAL_PLACE = "04"


class IndicatorData:
    pass


class IndicatorOperator:
    @staticmethod
    def get_column_names():
        pass

    @staticmethod
    def calculate_final_column_name(prefix, col_name, period):
        return f"{prefix}-{col_name}-{period}"


    def apply(self, data: IndicatorData) -> IndicatorData:
        pass


###############
# tech indicators
###############

class RsiOperator(IndicatorOperator):

    @staticmethod
    def get_column_names():
        return ["Rsi"]

    def __init__(self, prefix, period):
        self.prefix = prefix
        self.period = period
        self.logger = logging.getLogger(__name__)

    def apply(self, data: IndicatorData):
        self.logger.info("calculating rsi =================================")
        col_name = self.get_column_names()[0]
        final_col_name = self.calculate_final_column_name(prefix=self.prefix, col_name=col_name, period=self.period)
        data[final_col_name] = RSI(data["Close"], timeperiod=self.period)
        return data


class BollingerBandOperator(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Bb-high", "Bb-low"]

    def __init__(self, prefix, period):
        self.logger = logging.getLogger(__name__)
        self.period = period
        self.prefix = prefix
        pass

    def apply(self, data: IndicatorData) -> IndicatorData:
        self.logger.info("calculating bollinger band =================================")
        bb_high_name = self.get_column_names()[0]
        bb_low_name = self.get_column_names()[1]

        bband_df = self._compute_bb(data["Close"],
                                    look_back_period=self.period,
                                    prefix=self.prefix,
                                    bb_low_name=bb_low_name,
                                    bb_high_name=bb_high_name)
        data = data.join(bband_df)

        bb_high_name_full = self.calculate_final_column_name(prefix=self.prefix, col_name=bb_high_name, period=self.period)
        bb_low_name_full = self.calculate_final_column_name(prefix=self.prefix, col_name=bb_low_name, period=self.period)


        data[bb_high_name_full] = data[
            bb_high_name_full].sub(
            data["Close"]).div(data[bb_high_name_full]).apply(
            np.log1p)
        data[bb_low_name_full] = data["Close"].sub(
            data[bb_low_name_full]).div(data["Close"]).apply(
            np.log1p)

        return data

    def _compute_bb(self, price, look_back_period, prefix, bb_high_name, bb_low_name):
        """
            - high and low is just variances bollinger band
        """
        high, mid, low = BBANDS(price, timeperiod=look_back_period)
        return pd.DataFrame({self.calculate_final_column_name(prefix=self.prefix, col_name=bb_high_name, period=self.period): high,
                             self.calculate_final_column_name(prefix=self.prefix, col_name=bb_low_name, period=self.period): low})


class AtrOperator(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Natr", "Atr"]

    def __init__(self, prefix, period):
        self.prefix = prefix
        self.period = period
        self.logger = logging.getLogger(__name__)

    def apply(self, data: IndicatorData) -> IndicatorData:
        self.logger.info("Average true range =================================")
        natr_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        atr_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[1], period=self.period)
        ## Natr: Normalized Average True Range
        data[natr_col] = NATR(data["High"],
                              data["Low"],
                              data["Close"],
                              timeperiod=self.period)

        atr_df = ATR(data["High"], data["Low"], data["Close"], timeperiod=self.period)
        # todo: removed cause it hints the model about the mean and std of the whole train data!!!
        # ohlc_price[f"{result_col_prefix}Atr_{look_back_period:{DECIMAL_PLACE}}"] = atr_df.sub(atr_df.mean()).div(atr_df.std())
        data[atr_col] = atr_df

        return data


class MaOperator(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Ppo", "Macd"]

    def __init__(self, prefix, period):
        self.prefix = prefix
        self.period = period
        self.logger = logging.getLogger(__name__)

    def apply(self, data: IndicatorData) -> IndicatorData:
        self.logger.info("Moving Average Convergence/Divergence =================================")
        ppo_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        macd_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[1], period=self.period)

        ppo_df = PPO(data["Close"],
                     fastperiod=self.period,
                     matype=1)
        data[ppo_col] = ppo_df

        # todo: remove it
        # ppo_df.to_hdf("tmp_data.h5", "f_ppo")
        # ohlc_price.to_hdf("tmp_data.h5", "f_close")

        macd_df = MACD(data["Close"], signalperiod=self.period)[0]
        data[macd_col] = macd_df

        return data


class VilliamrOperator(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Willr"]

    def __init__(self, prefix, period):
        self.prefix = prefix
        self.period = period
        self.logger = logging.getLogger(__name__)

    def apply(self, data: IndicatorData) -> IndicatorData:
        willr_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        data[willr_col] = WILLR(data["High"],
                                                                                data["Low"],
                                                                                data["Close"],
                                                                                timeperiod=self.period)
        return data


class MomentomOperator(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Mom"]

    def __init__(self, prefix, period):
        self.prefix = prefix
        self.period = period
        self.logger = logging.getLogger(__name__)

    def apply(self, data: IndicatorData) -> IndicatorData:
        mom_col =self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        data[mom_col] = MOM(data["Close"],
                                                                            timeperiod=self.period)

        return data


class WmaOperator(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Wma"]

    def __init__(self, prefix, period):
        self.prefix = prefix
        self.period = period
        self.logger = logging.getLogger(__name__)

    def apply(self, data: IndicatorData) -> IndicatorData:
        wma_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        data[wma_col] = WMA(data["Close"], timeperiod=self.period)

        return data


class EmaOperator(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Ema"]

    def __init__(self, prefix, period):
        self.prefix = prefix
        self.period = period
        self.logger = logging.getLogger(__name__)

    def apply(self, data: IndicatorData) -> IndicatorData:
        ema_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        data[ema_col] = EMA(data["Close"],
                                                                            timeperiod=self.period)

        return data

class SmaOperator(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Sma"]

    def __init__(self, prefix, period):
        self.prefix = prefix
        self.period = period
        self.logger = logging.getLogger(__name__)

    def apply(self, data: IndicatorData) -> IndicatorData:
        ema_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        data[ema_col] = SMA(data["Close"], timeperiod=self.period)

        return data


class AdxOperator(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Adx"]

    def __init__(self, prefix, period):
        self.prefix = prefix
        self.period = period
        self.logger = logging.getLogger(__name__)

    def apply(self, data: IndicatorData) -> IndicatorData:
        adx_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        data[adx_col] = ADX(data["High"],
                            data["Low"],
                            data["Close"],
                            timeperiod=self.period)

        return data


#########################
# tech indicator: intraday
# todo: needs more work:
#  https://github.com/PacktPublishing/Machine-Learning-for-Algorithmic-Trading-Second-Edition/blob/master/12_gradient_boosting_machines/10_intraday_features.ipynb
#############################
class BopOperator(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Bop"]

    def __init__(self, prefix, period):
        self.prefix = prefix
        self.period = period
        self.logger = logging.getLogger(__name__)

    def apply(self, data: IndicatorData):
        self.logger.info("calculating bop =================================")
        col_name = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        data[col_name] = talib.BOP(
            data["Open"],
            data["High"],
            data["Low"],
            data["Close"])
        return data


###############
# returns
###############

class ReturnsOperator(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Returns"]

    def __init__(self, prefix, period, magnifier_multiplier=10000):
        self.prefix = prefix
        self.period = period
        self.magnifier_multiplier = magnifier_multiplier

    def apply(self, data: IndicatorData) -> IndicatorData:
        logger = logging.getLogger(__name__)
        logger.info("Historical Returns =================================")
        result_col_prefix = self.prefix
        close_prices = data["Close"]

        hist_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        # Todo = 1 for now but returns so small
        data[hist_col] = close_prices.pct_change(
            self.period) * self.magnifier_multiplier
        return data


###############
# max min operators
###############

class MaxPriceAndReturn(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Price-Max", "Return-Max"]

    def __init__(self, prefix, period):
        self.prefix = prefix
        self.period = period

    def apply(self, data: IndicatorData) -> IndicatorData:
        max_price = data["Close"].rolling(self.period).max()
        current_price = data["Close"]

        price_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        return_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[1], period=self.period)

        data[return_col] = (max_price - current_price).div(
            current_price)

        data[price_col] = max_price
        return data


class MinPriceAndReturn(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Price-Min", "Return-Min"]

    def __init__(self, prefix, period):
        self.prefix = prefix
        self.period = period

    def apply(self, data: IndicatorData) -> IndicatorData:
        min_price = data["Close"].rolling(self.period).min()
        current_price = data["Close"]

        price_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        return_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[1], period=self.period)

        data[return_col] = (min_price - current_price).div(
            current_price)

        data[price_col] = min_price
        return data


class ForwardHighestReturn(IndicatorOperator):
    """
        - todo: so slow cos of if-else
        - returns highest value return with sign (can be from min (neg) or max (pos)
    """

    @staticmethod
    def get_column_names():
        return ["Return-Highest"]

    def __init__(self, prefix, prefix_max_return, prefix_min_return, period):
        self.prefix = prefix
        self.prefix_max_return = prefix_max_return
        self.prefix_min_return = prefix_min_return
        self.period = period

    def apply(self, data: IndicatorData) -> IndicatorData:
        min_forward = f"{self.prefix_min_return}"
        max_forward = f"{self.prefix_max_return}"
        high_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)

        data["temp_1"] = data[max_forward].abs() - data[min_forward].abs()
        data[high_col] = np.where(
            data["temp_1"] > 0, data[max_forward], data[min_forward])
        _ = data.pop("temp_1")

        return data

#################
# return volatitliy
##################

class ReturnVarianceOperator(IndicatorOperator):
    @staticmethod
    def get_column_names():
        return ["Return-Var"]

    def __init__(self, prefix, period, magnifier_multiplier=1):
        self.prefix = prefix
        self.period = period
        self.magnifier_multiplier = magnifier_multiplier

    def apply(self, data: IndicatorData) -> IndicatorData:
        logger = logging.getLogger(__name__)
        logger.info("Returns variances =================================")
        result_col_prefix = self.prefix
        close_prices = data["Close"]

        res_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0], period=self.period)
        # Todo magnifier= 1 for now but returns so small
        # Todo: pct change fore return can be params
        data["_temp"] = close_prices.pct_change(1) * self.magnifier_multiplier
        data[res_col] = data["_temp"].rolling(self.period).std()
        _ = data.pop("_temp")
        return data



##############
# lags
##############
class Lag(IndicatorOperator):
    """
        - todo: so slow cos of if-else
        - returns highest value return with sign (can be from min (neg) or max (pos)
    """

    @staticmethod
    def get_column_names():
        return ["T"]

    def __init__(self, prefix, prefix_operator, period):
        """
        :param prefix: master prefix
        :param prefix_operator: operator prefix to apply lag to
        :param period: period that lag need to apply
        """
        self.prefix = prefix
        self.prefix_operator = prefix_operator
        self.period = period

    def apply(self, data: IndicatorData) -> IndicatorData:
        prefix_operator = f"{self.prefix_operator}"
        lag_col = self.get_column_names()[0]

        col_to_lag = f"{prefix_operator}"

        # high_col = self.calculate_final_column_name(prefix=self.prefix, col_name=self.get_column_names()[0],
        #                                             period=self.period)
        data[f"{self.prefix}-{prefix_operator}-{lag_col}-{self.period:{DECIMAL_PLACE}}"] = data[col_to_lag].shift(
            self.period)

        return data


##############
# others
##############

class CandleStickOperators(IndicatorOperator):

    @staticmethod
    def get_column_names():
        return ["CStick_Body", "Top_Wick", "Bottom_Wick"]

    def __init__(self, prefix):
        self.prefix = prefix

    def apply(self, data: IndicatorData) -> IndicatorData:
        logger = logging.getLogger(__name__)
        logger.info("candle sticks =================================")
        body_col, top_col, bottom_col = self.get_column_names()

        data[f"{self.prefix}{body_col}"] = data["Open"] - data["Close"]
        data[f"{self.prefix}{top_col}"] = data.apply(lambda x: self._calculate_top_wick(x, body_col=body_col), axis=1)
        data[f"{self.prefix}{bottom_col}"] = data.apply(lambda x: self._calculate_top_wick(x, body_col=body_col),
                                                        axis=1)

    @staticmethod
    def _calculate_top_wick(row, body_col):
        if row[body_col] > 0:
            return row["High"] - row["Open"]
        else:
            return row["High"] - row["Close"]

    @staticmethod
    def _calculate_bottom_wick(row, body_col):
        if row[body_col] > 0:
            return row["Close"] - row["Low"]
        else:
            return row["Open"] - row["Low"]


###############
# tseries_patterns indicators
###############
# from tseries_patterns import AmplitudeBasedLabeler
#
#
# class AmplitudeBasedLabelerOperator(IndicatorOperator):
#     @staticmethod
#     def get_column_names():
#         return ["Amp"]
#
#     def __init__(self, prefix, period):
#         self.prefix = prefix
#         self.logger = logging.getLogger(__name__)
#         # todo: hard coded
#         self.minamp = 5
#         self.tinactive = period
#         self.period = period
#         self.labeler = AmplitudeBasedLabeler(minamp=self.minamp, Tinactive=self.tinactive)
#
#     def apply(self, data: IndicatorData):
#         self.logger.info("calculating rsi =================================")
#         col_name = self.get_column_names()[0]
#         orig_indx_name = data.index.name
#         data.index.name = "stamp"
#
#         labeled = self.labeler.label(data[["Close"]])
#         labeled = labeled.set_index("stamp")
#
#         data[f"{self.prefix}{col_name}_{self.period:{DECIMAL_PLACE}}"] = labeled["label"].shift(-self.period)
#
#         data.index.name = orig_indx_name
#         return data



"""
- todo:
    - sensitive to scaling, noise, outliers, ...
    - just for regression (none for classifiers)
"""

import pandas as pd
from sklearn.feature_selection import *
from sklearn.ensemble import RandomForestRegressor

RANDOM_FOREST_ESTIMATORS_N = 5
RANDOM_FOREST_DEPTH = 3


def remove_low_variance_features(features: pd.DataFrame):
    # todo: it is fixed, sensitive to scaling, ...
    # todo: how it works per feature?
    threshold = 0.01
    sel = VarianceThreshold(threshold=threshold)
    _ = sel.fit_transform(features)
    return features[features.columns[sel.get_support()]]


def remove_weak_linear(features: pd.DataFrame, targets: pd.DataFrame, percentile_to_keep=80):
    sp = SelectPercentile(f_regression, percentile=percentile_to_keep)
    sp.fit_transform(features, targets)
    return features[features.columns[sp.get_support()]]


def remove_features_recursively(features: pd.DataFrame, targets: pd.DataFrame, n_features_to_select: int):
    estimator = RandomForestRegressor(n_estimators=RANDOM_FOREST_ESTIMATORS_N, max_depth=RANDOM_FOREST_DEPTH)
    rfe = RFE(estimator=estimator, n_features_to_select=n_features_to_select, step=1)
    rfe.fit(features.values, targets.values.ravel())
    return features[features.columns[rfe.get_support()]]


def remove_features_using_model(features: pd.DataFrame, targets: pd.DataFrame):
    estimator = RandomForestRegressor(n_estimators=RANDOM_FOREST_ESTIMATORS_N, max_depth=RANDOM_FOREST_DEPTH)
    sfm = SelectFromModel(estimator=estimator, threshold=0)
    sfm.fit(features.values, targets.values.ravel())
    # importance
    return features[features.columns[sfm.get_support()]]


def remove_features_sequentially(features: pd.DataFrame, targets: pd.DataFrame, n_features_to_select: int):
    estimator = RandomForestRegressor(n_estimators=RANDOM_FOREST_ESTIMATORS_N, max_depth=RANDOM_FOREST_DEPTH)

    seqfm = SequentialFeatureSelector(estimator=estimator,
                                      n_features_to_select=n_features_to_select,
                                      direction="forward")
    seqfm.fit(features.values, targets.values.ravel())

    return features[features.columns[seqfm.get_support()]]



import pandas as pd
import pytest
import numpy as np

from moosir_feature.trades.alpha_manager import create_quantile_alphas, create_absolute_prediction_alphas


def test_create_quantile_alphas():
    # arrange

    preds_df = pd.DataFrame(data={"pred_col": [0.1, 0.2, 0.3, 0.4, 0.5]})
    instances = pd.DataFrame(data=np.random.random([5, 1]))

    # assert
    result = create_quantile_alphas(instances=instances, prediction_result=preds_df, quantile_threshold=0.2)

    # act
    assert result["Signal"].iloc[0] == -1
    assert result["Signal"].iloc[1] == 0
    assert result["Signal"].iloc[2] == 0
    assert result["Signal"].iloc[3] == 0
    assert result["Signal"].iloc[4] == 1


def test_create_absolute_prediction_alphas():
    # arrange
    preds_df = pd.DataFrame(data={"pred_col": [1, 2, 3, 4, 0.5]})
    instances = pd.DataFrame(data=np.random.random([5, 1]))

    # assert
    result = create_absolute_prediction_alphas(instances=instances, prediction_result=preds_df)

    # act
    assert result["Signal"].iloc[0] == 1
    assert result["Signal"].iloc[1] == 2
    assert result["Signal"].iloc[2] == 3
    assert result["Signal"].iloc[3] == 4
    assert result["Signal"].iloc[4] == 0.5



from moosir_feature.model_validations.benchmarking import RandomModel
from moosir_feature.model_validations.model_cv_runner import *

import pytest
import pandas as pd
import numpy as np
from moosir_feature.model_validations.model_validator import CustomTsCv
from sklearn.dummy import DummyRegressor


def test_predict_on_cv():

    # arrange
    sample_n = 100
    feature_n = 1

    feature_cols = [f"f_{i}" for i in range(feature_n)]
    target_col = "target"
    features = pd.DataFrame(data=np.arange(0, sample_n * feature_n).reshape(sample_n, -1),
                            columns=feature_cols,
                            index=pd.date_range("2000/01/01", periods=sample_n))
    targets = pd.DataFrame(data=np.arange(0, sample_n * 1).reshape(sample_n, -1),
                           columns=[target_col],
                           index=pd.date_range("2000/01/01", periods=sample_n))

    model = DummyRegressor(strategy="mean")
    cv = CustomTsCv(train_n=10, test_n=3, sample_n=sample_n, train_shuffle_block_size=None)

    # act
    result = predict_on_cv(model=model, features=features, targets=targets, cv=cv)

    # assert
    assert len(result) == sample_n
    assert len(result.columns) == len(result.columns)






